/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "bindinterface.h"
#include "audioengine.h"
#include "audioport.h"
#include "basicclock.h"
#include "beattracker.h"
#include "bindaudio.h"
#include "bindio.h"
#include "bindjack.h"
#include "bindlv2.h"
#include "bindmidi.h"
#include "bindscript.h"
#include "bindtime.h"
#include "io.h"
#include "jackclock.h"
#include "link.h"
#include "lv2plugin.h"
#include "midibeatclock.h"
#include "midifile.h"
#include "midiport.h"
#include "mixer.h"
#include "onsetdetector.h"
#include "timebase.h"
#include "variableclock.h"
#include "zipfile.h"

namespace bipscript {
namespace binding {

audio::Source *getAudioSource(HSQUIRRELVM &vm, int index) {
    SQUserPointer ptr;
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &AudioMixerObject, false))) {
        return static_cast<audio::Mixer *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &AudioMultiInputObject, false))) {
        return static_cast<audio::MultiInput *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &AudioInputObject, false))) {
        return static_cast<audio::AudioInputPort *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &AudioStereoInputObject, false))) {
        return static_cast<audio::StereoInput *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &Lv2PluginObject, false))) {
        return static_cast<lv2::Plugin *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, nullptr, false))) {
        return resolve_audiosource(ptr);
    }
	return nullptr;
}

audio::Source *getAudioSource(SQUserPointer type, HSQOBJECT *arg) {
    if(type == &AudioMixerObject) {
        return static_cast<audio::Mixer *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &AudioMultiInputObject) {
        return static_cast<audio::MultiInput *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &AudioInputObject) {
        return static_cast<audio::AudioInputPort *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &AudioStereoInputObject) {
        return static_cast<audio::StereoInput *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &Lv2PluginObject) {
        return static_cast<lv2::Plugin *>(bip_getinstanceuserpointer(arg));
    }
    return nullptr;
}

audio::Sink *getAudioSink(HSQUIRRELVM &vm, int index) {
    SQUserPointer ptr;
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &AudioMixerObject, false))) {
        return static_cast<audio::Mixer *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &AudioMultiOutputObject, false))) {
        return static_cast<audio::MultiOutput *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &AudioOnsetDetectorObject, false))) {
        return static_cast<audio::OnsetDetector *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &AudioOutputObject, false))) {
        return static_cast<audio::Output *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &AudioStereoOutputObject, false))) {
        return static_cast<audio::StereoOutput *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &Lv2PluginObject, false))) {
        return static_cast<lv2::Plugin *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, nullptr, false))) {
        return resolve_audiosink(ptr);
    }
	return nullptr;
}

audio::Sink *getAudioSink(SQUserPointer type, HSQOBJECT *arg) {
    if(type == &AudioMixerObject) {
        return static_cast<audio::Mixer *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &AudioMultiOutputObject) {
        return static_cast<audio::MultiOutput *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &AudioOnsetDetectorObject) {
        return static_cast<audio::OnsetDetector *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &AudioOutputObject) {
        return static_cast<audio::Output *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &AudioStereoOutputObject) {
        return static_cast<audio::StereoOutput *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &Lv2PluginObject) {
        return static_cast<lv2::Plugin *>(bip_getinstanceuserpointer(arg));
    }
    return nullptr;
}

io::Path *getIOPath(HSQUIRRELVM &vm, int index) {
    SQUserPointer ptr;
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &IOFileObject, false))) {
        return static_cast<io::File *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &IOFolderObject, false))) {
        return static_cast<io::Folder *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &IOZipFileObject, false))) {
        return static_cast<io::ZipFile *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &IOZipFolderObject, false))) {
        return static_cast<io::ZipFolder *>(ptr);
    }
	return nullptr;
}

io::Path *getIOPath(SQUserPointer type, HSQOBJECT *arg) {
    if(type == &IOFileObject) {
        return static_cast<io::File *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &IOFolderObject) {
        return static_cast<io::Folder *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &IOZipFileObject) {
        return static_cast<io::ZipFile *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &IOZipFolderObject) {
        return static_cast<io::ZipFolder *>(bip_getinstanceuserpointer(arg));
    }
    return nullptr;
}

midi::Message *getMidiMessage(HSQUIRRELVM &vm, int index) {
    SQUserPointer ptr;
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiCopyrightObject, false))) {
        return static_cast<midi::Copyright *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiCuePointObject, false))) {
        return static_cast<midi::CuePoint *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiInstrumentObject, false))) {
        return static_cast<midi::Instrument *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiLyricObject, false))) {
        return static_cast<midi::Lyric *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiMarkerObject, false))) {
        return static_cast<midi::Marker *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiNoteOnObject, false))) {
        return static_cast<midi::NoteOn *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiNoteOffObject, false))) {
        return static_cast<midi::NoteOff *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiKeySignatureObject, false))) {
        return static_cast<midi::KeySignature *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiControlObject, false))) {
        return static_cast<midi::Control *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiTempoChangeObject, false))) {
        return static_cast<midi::TempoChange *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiTextObject, false))) {
        return static_cast<midi::Text *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiTimeSignatureObject, false))) {
        return static_cast<midi::TimeSignature *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiTrackNameObject, false))) {
        return static_cast<midi::TrackName *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiPitchBendObject, false))) {
        return static_cast<midi::PitchBend *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiProgramChangeObject, false))) {
        return static_cast<midi::ProgramChange *>(ptr);
    }
	return nullptr;
}

midi::Message *getMidiMessage(SQUserPointer type, HSQOBJECT *arg) {
    if(type == &MidiCopyrightObject) {
        return static_cast<midi::Copyright *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiCuePointObject) {
        return static_cast<midi::CuePoint *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiInstrumentObject) {
        return static_cast<midi::Instrument *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiLyricObject) {
        return static_cast<midi::Lyric *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiMarkerObject) {
        return static_cast<midi::Marker *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiNoteOnObject) {
        return static_cast<midi::NoteOn *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiNoteOffObject) {
        return static_cast<midi::NoteOff *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiKeySignatureObject) {
        return static_cast<midi::KeySignature *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiControlObject) {
        return static_cast<midi::Control *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiTempoChangeObject) {
        return static_cast<midi::TempoChange *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiTextObject) {
        return static_cast<midi::Text *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiTimeSignatureObject) {
        return static_cast<midi::TimeSignature *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiTrackNameObject) {
        return static_cast<midi::TrackName *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiPitchBendObject) {
        return static_cast<midi::PitchBend *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiProgramChangeObject) {
        return static_cast<midi::ProgramChange *>(bip_getinstanceuserpointer(arg));
    }
    return nullptr;
}

midi::Source *getMidiSource(HSQUIRRELVM &vm, int index) {
    SQUserPointer ptr;
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &Lv2PluginObject, false))) {
        return static_cast<lv2::Plugin *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiClockGeneratorObject, false))) {
        return static_cast<midi::ClockGenerator *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiInputObject, false))) {
        return static_cast<midi::MidiInputPort *>(ptr);
    }
	return nullptr;
}

midi::Source *getMidiSource(SQUserPointer type, HSQOBJECT *arg) {
    if(type == &Lv2PluginObject) {
        return static_cast<lv2::Plugin *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiClockGeneratorObject) {
        return static_cast<midi::ClockGenerator *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiInputObject) {
        return static_cast<midi::MidiInputPort *>(bip_getinstanceuserpointer(arg));
    }
    return nullptr;
}

midi::Sink *getMidiSink(HSQUIRRELVM &vm, int index) {
    SQUserPointer ptr;
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &Lv2PluginObject, false))) {
        return static_cast<lv2::Plugin *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiOutputObject, false))) {
        return static_cast<midi::MidiOutputPort *>(ptr);
    }
	return nullptr;
}

midi::Sink *getMidiSink(SQUserPointer type, HSQOBJECT *arg) {
    if(type == &Lv2PluginObject) {
        return static_cast<lv2::Plugin *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiOutputObject) {
        return static_cast<midi::MidiOutputPort *>(bip_getinstanceuserpointer(arg));
    }
    return nullptr;
}

time::Clock *getTimeClock(HSQUIRRELVM &vm, int index) {
    SQUserPointer ptr;
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &AudioBeatTrackerObject, false))) {
        return static_cast<audio::BeatTracker *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &JackClockObject, false))) {
        return static_cast<jack::Clock *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &JackTimebaseObject, false))) {
        return static_cast<jack::Timebase *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &MidiBeatTrackerObject, false))) {
        return static_cast<midi::BeatTracker *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &TimeBasicClockObject, false))) {
        return static_cast<time::BasicClock *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &TimeLinkObject, false))) {
        return static_cast<time::Link *>(ptr);
    }
    if (!SQ_FAILED(sq_getinstanceup(vm, index, &ptr, &TimeVariableClockObject, false))) {
        return static_cast<time::VariableClock *>(ptr);
    }
	return nullptr;
}

time::Clock *getTimeClock(SQUserPointer type, HSQOBJECT *arg) {
    if(type == &AudioBeatTrackerObject) {
        return static_cast<audio::BeatTracker *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &JackClockObject) {
        return static_cast<jack::Clock *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &JackTimebaseObject) {
        return static_cast<jack::Timebase *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &MidiBeatTrackerObject) {
        return static_cast<midi::BeatTracker *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &TimeBasicClockObject) {
        return static_cast<time::BasicClock *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &TimeLinkObject) {
        return static_cast<time::Link *>(bip_getinstanceuserpointer(arg));
    }
    if(type == &TimeVariableClockObject) {
        return static_cast<time::VariableClock *>(bip_getinstanceuserpointer(arg));
    }
    return nullptr;
}

void pushIOPath(HSQUIRRELVM &vm, io::Path *obj) {
    if (io::File* ret = dynamic_cast<io::File *>(obj)) {
        sq_pushobject(vm, IOFileObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
        sq_setreleasehook(vm, -1, BindUtil::Delete<io::File>);
    }
    if (io::Folder* ret = dynamic_cast<io::Folder *>(obj)) {
        sq_pushobject(vm, IOFolderObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
        sq_setreleasehook(vm, -1, BindUtil::Delete<io::Folder>);
    }
    if (io::ZipFile* ret = dynamic_cast<io::ZipFile *>(obj)) {
        sq_pushobject(vm, IOZipFileObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
        sq_setreleasehook(vm, -1, BindUtil::Delete<io::ZipFile>);
    }
    if (io::ZipFolder* ret = dynamic_cast<io::ZipFolder *>(obj)) {
        sq_pushobject(vm, IOZipFolderObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
        sq_setreleasehook(vm, -1, BindUtil::Delete<io::ZipFolder>);
    }
}
void pushMidiMessage(HSQUIRRELVM &vm, midi::Message *obj) {
    if (midi::Copyright* ret = dynamic_cast<midi::Copyright *>(obj)) {
        sq_pushobject(vm, MidiCopyrightObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::Copyright>);
    }
    if (midi::CuePoint* ret = dynamic_cast<midi::CuePoint *>(obj)) {
        sq_pushobject(vm, MidiCuePointObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::CuePoint>);
    }
    if (midi::Instrument* ret = dynamic_cast<midi::Instrument *>(obj)) {
        sq_pushobject(vm, MidiInstrumentObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::Instrument>);
    }
    if (midi::Lyric* ret = dynamic_cast<midi::Lyric *>(obj)) {
        sq_pushobject(vm, MidiLyricObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::Lyric>);
    }
    if (midi::Marker* ret = dynamic_cast<midi::Marker *>(obj)) {
        sq_pushobject(vm, MidiMarkerObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::Marker>);
    }
    if (midi::NoteOn* ret = dynamic_cast<midi::NoteOn *>(obj)) {
        sq_pushobject(vm, MidiNoteOnObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::NoteOn>);
    }
    if (midi::NoteOff* ret = dynamic_cast<midi::NoteOff *>(obj)) {
        sq_pushobject(vm, MidiNoteOffObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::NoteOff>);
    }
    if (midi::KeySignature* ret = dynamic_cast<midi::KeySignature *>(obj)) {
        sq_pushobject(vm, MidiKeySignatureObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::KeySignature>);
    }
    if (midi::Control* ret = dynamic_cast<midi::Control *>(obj)) {
        sq_pushobject(vm, MidiControlObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::Control>);
    }
    if (midi::TempoChange* ret = dynamic_cast<midi::TempoChange *>(obj)) {
        sq_pushobject(vm, MidiTempoChangeObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::TempoChange>);
    }
    if (midi::Text* ret = dynamic_cast<midi::Text *>(obj)) {
        sq_pushobject(vm, MidiTextObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::Text>);
    }
    if (midi::TimeSignature* ret = dynamic_cast<midi::TimeSignature *>(obj)) {
        sq_pushobject(vm, MidiTimeSignatureObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::TimeSignature>);
    }
    if (midi::TrackName* ret = dynamic_cast<midi::TrackName *>(obj)) {
        sq_pushobject(vm, MidiTrackNameObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::TrackName>);
    }
    if (midi::PitchBend* ret = dynamic_cast<midi::PitchBend *>(obj)) {
        sq_pushobject(vm, MidiPitchBendObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::PitchBend>);
    }
    if (midi::ProgramChange* ret = dynamic_cast<midi::ProgramChange *>(obj)) {
        sq_pushobject(vm, MidiProgramChangeObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret->addref());
        sq_setreleasehook(vm, -1, BindUtil::Decrement<midi::ProgramChange>);
    }
}
void pushTimeClock(HSQUIRRELVM &vm, time::Clock *obj) {
    if (audio::BeatTracker* ret = dynamic_cast<audio::BeatTracker *>(obj)) {
        sq_pushobject(vm, AudioBeatTrackerObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
    }
    if (jack::Clock* ret = dynamic_cast<jack::Clock *>(obj)) {
        sq_pushobject(vm, JackClockObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
    }
    if (jack::Timebase* ret = dynamic_cast<jack::Timebase *>(obj)) {
        sq_pushobject(vm, JackTimebaseObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
    }
    if (midi::BeatTracker* ret = dynamic_cast<midi::BeatTracker *>(obj)) {
        sq_pushobject(vm, MidiBeatTrackerObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
    }
    if (time::BasicClock* ret = dynamic_cast<time::BasicClock *>(obj)) {
        sq_pushobject(vm, TimeBasicClockObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
    }
    if (time::Link* ret = dynamic_cast<time::Link *>(obj)) {
        sq_pushobject(vm, TimeLinkObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
    }
    if (time::VariableClock* ret = dynamic_cast<time::VariableClock *>(obj)) {
        sq_pushobject(vm, TimeVariableClockObject);
        sq_createinstance(vm, -1);
        sq_remove(vm, -2);
        sq_setinstanceup(vm, -1, ret);
    }
}

}
}

