/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "bipxtapi.h"

#include "bipsquirrel.h"
#include "extension.h"
#include "xtfunctioncall.h"
#include "xtprocessor.h"

#include <stdio.h>
#include <string.h>

HSQUIRRELVM vm;

using namespace bipscript;

void bipxt_setvm(HSQUIRRELVM ivm) { vm = ivm; }

// timeline
BXTTimeline bipxt_timeline_create() { return new XtTimelineSet(); }

void bipxt_timeline_schedule(BXTTimeline xtimeline, void *message, float measure) {
    XtTimelineSet *timeline = static_cast<XtTimelineSet *>(xtimeline);
    timeline->createEvent(new XtMessage(message), measure, Clock::defaultClock());
}

void *bipxt_timeline_event_first(BXTTimeline xtimeline) {
    XtTimelineSet *timeline = static_cast<XtTimelineSet *>(xtimeline);
    Event<XtMessage> *evt = timeline->firstEvent();
    return evt ? evt->message().mdata : nullptr;
}

void *bipxt_timeline_event_next(BXTTimeline xtimeline) {
    XtTimelineSet *timeline = static_cast<XtTimelineSet *>(xtimeline);
    Event<XtMessage> *evt = timeline->nextEvent();
    return evt ? evt->message().mdata : nullptr;
}

int bipxt_timeline_active(BXTTimeline xtimeline) {
    return static_cast<XtTimelineSet *>(xtimeline)->active();
}

// process thread
BXTProcessThread bipxt_create_thread(BXTProcess process, BXTProcess restart, void *arg) {
    ProcessorEngine *engine = new XtProcessor(process, restart, arg);
    ScriptHost::instance().addProcessorThread(engine);
    return engine;
}

BXTClass bipxt_bind_class(const char *name, uint32_t namelen, BXTMethod ctor) {
    HSQOBJECT object;
    sq_pushstring(vm, _SC(name), namelen);
    sq_newclass(vm, false);
    sq_getstackobj(vm, -1, &object);
    sq_settypetag(vm, -1, object._unVal.pClass);
    bipxt_bind_method("constructor", 11, ctor);
    return object._unVal.pClass;
}

void bipxt_bind_class_end() { sq_newslot(vm, -3, false); }

void bipxt_bind_enumeration(const char *name) {
    sq_pushstring(vm, name, -1);
    sq_newtable(vm);
}

void bipxt_bind_enumerator(const char *name, int32_t value) {
    sq_pushstring(vm, name, -1);
    sq_pushinteger(vm, value);
    sq_newslot(vm, -3, true);
}

void bipxt_bind_enumeration_end() { sq_newslot(vm, -3, true); }

SQRESULT exec_method(HSQUIRRELVM vm) {
    BXTMethod *method;
    sq_getuserdata(vm, -1, (SQUserPointer *)&method, NULL);
    return (*method)(sq_gettop(vm) - 1);
}

void bipxt_bind_method(const char *name, uint32_t namelen, BXTMethod func) {
    sq_pushstring(vm, _SC(name), namelen);
    size_t size = sizeof(func);
    SQUserPointer udata = sq_newuserdata(vm, static_cast<SQUnsignedInteger>(size));
    memcpy(udata, &func, size);
    sq_newclosure(vm, exec_method, 1);
    sq_newslot(vm, -3, false);
}

BXTResult bipxt_get_boolean(BXTIndex index, BXTBoolean *boolp) {
    SQBool sqbool;
    BXTResult result = sq_getbool(vm, index, &sqbool);
    *boolp = sqbool;
    return result;
}

BXTResult bipxt_get_float(BXTIndex index, BXTFloat *floatp) {
    return sq_getfloat(vm, index, floatp);
}

void bipxt_put_float(float value) { sq_pushfloat(vm, value); }

BXTResult bipxt_get_integer(BXTIndex index, BXTInteger *intp) {
    return sq_getinteger(vm, index, intp);
}

void bipxt_put_integer(BXTInteger value) { sq_pushinteger(vm, value); }

BXTResult bipxt_get_string(BXTIndex index, const char **strp) {
    return sq_getstring(vm, index, strp);
}

void bipxt_put_string(const char *str, uint32_t strlen) { sq_pushstring(vm, str, strlen); }

// TODO: replace rdelete with function pointer to release hook
void bipxt_put_this(const void *obj, uint32_t rdelete) {
    sq_setinstanceup(vm, 1, (SQUserPointer *)obj);
}

BXTResult bipxt_get_this(void **ptr) { return sq_getinstanceup(vm, 1, ptr, 0, false); }

void bipxt_put_object(void *obj, BXTClass type) {
    HSQOBJECT classObj;
    classObj._type = OT_CLASS;
    classObj._unVal.pClass = static_cast<SQClass *>(type);
    sq_pushobject(vm, classObj);
    sq_createinstance(vm, -1);
    sq_remove(vm, -2);
    sq_setinstanceup(vm, -1, obj);
}

BXTResult bipxt_get_object(BXTIndex index, void **ptr, BXTClass type) { // C++ object
    SQUserPointer objPtr;
    if (!SQ_FAILED(sq_getinstanceup(vm, index, (SQUserPointer *)&objPtr, type, false))) {
        *ptr = objPtr;
        return SQ_OK;
    }
    return SQ_ERROR;
}

BXTResult bipxt_get_instance(BXTIndex index, BXTInstance *ptr) {
    HSQOBJECT obj;
    sq_getstackobj(vm, index, &obj);
    if (obj._type != OT_INSTANCE) {
        return SQ_ERROR;
    }
    *ptr = obj._unVal.pInstance;
    return SQ_OK;
}

BXTResult bipxt_get_function(BXTIndex index, BXTFunction *ptr) {
    HSQOBJECT obj;
    sq_getstackobj(vm, index, &obj);
    if (obj._type != OT_CLOSURE) {
        return SQ_ERROR;
    }
    *ptr = obj._unVal.pClosure;
    return SQ_OK;
}

void bipxt_add_reference(BXTInstance object) {
    HSQOBJECT contextObj;
    contextObj._type = OT_INSTANCE;
    contextObj._unVal.pInstance = static_cast<SQInstance *>(object);
    sq_addref(vm, &contextObj);
}

BXTResult bipxt_execute(BXTFunction function, BXTInstance context) {
    HSQOBJECT funcObj;
    funcObj._type = OT_CLOSURE;
    funcObj._unVal.pClosure = static_cast<SQClosure *>(function);
    sq_pushobject(vm, funcObj);
    if (!context) {
        // push closure root as "this" pointer
        sq_getclosureroot(vm, -1);
    } else {
        HSQOBJECT contextObj;
        contextObj._type = OT_INSTANCE;
        contextObj._unVal.pInstance = static_cast<SQInstance *>(context);
        sq_pushobject(vm, contextObj);
    }
    // attempt to call function
    if (SQ_FAILED(sq_call(vm, 1, SQFalse, SQTrue))) {
        return SQ_ERROR;
    }
    // pop closure
    sq_pop(vm, 1);
    return SQ_OK;
}

BXTResult bipxt_error(const char *mesg) { return sq_throwerror(vm, mesg); }

BXTResult bipxt_not_found() {
    sq_pushnull(vm);
    return sq_throwobject(vm);
}

void bipxt_audio_processor(const char *name, BXTIndex inputs, BXTIndex outputs, void *obj,
                           BXTAudioProcess func) {
    bipscript::XtAudioProcessorMap::instance().add(obj, name, inputs, outputs, func);
}

BXTFloat *bipxt_audio_input_buffer(void *obj, BXTIndex index) {
    XtAudioProcessor *processor = XtAudioProcessorMap::instance().find(obj);
    audio::Connector &input = processor->inputbus()[index];
    audio::Connection *connection = input.connection();
    return connection ? connection->audio() : nullptr;
}

BXTFloat *bipxt_audio_output_buffer(void *obj, BXTIndex index) {
    XtAudioProcessor *processor = XtAudioProcessorMap::instance().find(obj);
    return processor->outputbus()[index].audio();
}

BXTFunctionCall bipxt_create_call(BXTFunction function) { return new XtFunctionCall(function); }

void bipxt_call_add_integer(BXTFunctionCall call, BXTInteger value) {
    XtFunctionCall *xtcall = static_cast<XtFunctionCall *>(call);
    xtcall->addInt(value);
}

BXTResult bipxt_execute_call(BXTFunctionCall call, BXTInstance context) {
    XtFunctionCall *xtcall = static_cast<XtFunctionCall *>(call);
    HSQOBJECT funcObj;
    funcObj._type = OT_CLOSURE;
    funcObj._unVal.pClosure = static_cast<SQClosure *>(xtcall->function());
    sq_pushobject(vm, funcObj);
    if (!context) {
        // push closure root as "this" pointer
        sq_getclosureroot(vm, -1);
    } else {
        HSQOBJECT contextObj;
        contextObj._type = OT_INSTANCE;
        contextObj._unVal.pInstance = static_cast<SQInstance *>(context);
        sq_pushobject(vm, contextObj);
    }
    // add args
    for (XtCallArg arg : xtcall->args()) {
        if (arg.type == XtCallArgType::INTEGER) {
            sq_pushinteger(vm, arg.ivalue);
        }
    }
    // attempt to call function
    if (SQ_FAILED(sq_call(vm, 1 + xtcall->nargs(), SQFalse, SQTrue))) {
        return SQ_ERROR;
    }
    // pop closure
    sq_pop(vm, 1);
    return SQ_OK;
}
