/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef SCRIPT_H
#define SCRIPT_H

#include "audioengine.h"
#include "bindscript.h"
#include "event.h"
#include "methodqueue.h"
#include "objectcache.h"
#include "scripthost.h"
#include "statusserver.h"
#include "timeline.h"
#include "virtualmachine.h"

namespace bipscript {
namespace script {

class StayAliveProcessor : public AudioProcessor {
  public:
    static StayAliveProcessor &instance() {
        static StayAliveProcessor instance;
        return instance;
    }
    void restart() {}
    bool doProcess(samples_t) { return true; }
};

class FunctionEvent;

class ScheduledClosure : public Closure,
                         public Listable<ScheduledClosure> { // replace with generic event closure
    FunctionEvent *mevent;

  protected:
    void addParameters() override {
        BindUtil::PushRef(vm, mevent, binding::ScriptFunctionEventObject);
    }

  public:
    ScheduledClosure(ScriptFunction &function) : Closure(function, nullptr) {}
    ScheduledClosure(ScriptFunction &function, SQInstance *context) : Closure(function, context) {}
    void event(FunctionEvent *event) { mevent = event; }
    void finished() override {}
};

class FunctionEvent : public Event<ScheduledClosure> {

  public:
    FunctionEvent(ScheduledClosure *closure, float measure, Clock &clock)
        : Event<ScheduledClosure>(closure, Event::Type::SCHEDULE, measure, clock) {}
};

class Script : public AudioProcessor, public ObjectCache {
    // command line arguments
    HSQOBJECT margv;      // handle to argv array
    HSQOBJECT marguments; // handle to arguments array
    // closure events
    bool mprocessing;
    TimelineSet<ScheduledClosure> eventBuffer;
    // singleton
    Script() : mprocessing(false) {}

  public:
    static void stayAlive() {
        AudioEngine::instance().addProcessor(&StayAliveProcessor::instance());
    }
    static Script &instance() {
        static Script instance;
        return instance;
    }
    static HSQOBJECT argv() { return instance().margv; }
    static HSQOBJECT argument() { return instance().marguments; }
    void init(int argc, char **argv, int fileindex) {
        HSQUIRRELVM vm = VirtualMachine::instance().vm();
        // create argv array
        sq_newarray(vm, argc);
        sq_getstackobj(vm, -1, &margv);
        sq_addref(vm, &margv);
        // add argv entries
        for (int i = 0; i < argc; i++) {
            sq_pushinteger(vm, i);
            sq_pushstring(vm, argv[i], -1);
            sq_set(vm, -3);
        }
        sq_poptop(vm);
        // create arguments array
        sq_newarray(vm, argc - fileindex);
        sq_getstackobj(vm, -1, &marguments);
        sq_addref(vm, &marguments);
        // add argument entries
        for (int i = fileindex; i < argc; i++) {
            sq_pushinteger(vm, i - fileindex);
            sq_pushstring(vm, argv[i], -1);
            sq_set(vm, -3);
        }
        sq_poptop(vm);
    }
    void exitScript(uint8_t code) {
        ScriptHost::instance().shutdown();
        if (AudioEngine::loaded) {
            AudioEngine::instance().shutdown();
        }
        if (StatusServer::instance().active()) {
            StatusServer::instance().shutdown();
        }
        CollectorThread::instance().shutdown();
        exit(code);
    }
    FunctionEvent *schedule(ScriptFunction &function, float measure,
                            Clock &clock = Clock::defaultClock()) {
        return schedule(function, nullptr, measure, clock);
    }
    FunctionEvent *schedule(ScriptFunction &function, SQInstance *context, float measure,
                            Clock &clock = Clock::defaultClock()) {
        validateFunction(function);
        if (!mprocessing) {
            AudioEngine::instance().addProcessor(this);
            mprocessing = true;
        }
        ScheduledClosure *closure =
            context ? new ScheduledClosure(function, context) : new ScheduledClosure(function);
        FunctionEvent *evt = new FunctionEvent(closure, measure, clock);
        closure->event(evt);
        eventBuffer.addEvent(evt, clock);
        reportEvent(&clock, evt, measure, closure);
        return evt;
    }
    void remove(FunctionEvent &evt) {
        eventBuffer.removeEvent(&evt, evt.clock());
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            statusserver.removeScriptEvent(&evt.clock(), &evt);
        }
    }
    bool doProcess(samples_t) override {
        Event<ScheduledClosure> *event = eventBuffer.firstEvent();
        bool active = false;
        while (event) {
            MethodQueue::instance().dispatch(&event->message());
            event = eventBuffer.nextEvent();
            active = true;
        }
        return active || eventBuffer.active();
    }
    // processor interface
    void restart() override { eventBuffer.restart(); }
    void restartScript() override { eventBuffer.restartScript(); }
    // objectcache interface
    void reset(bool) override { restartScript(); }
    void scriptComplete() override {}

  private:
    void reportEvent(Clock *clock, void *event, float measure, ScheduledClosure *closure) {
        if (StatusServer::instance().active()) {
            StatusServer::instance().addScriptEvent(clock, event, measure, closure->name());
        }
    }
    void validateFunction(ScriptFunction &function) {
        if (function.numargs() != 2) {
            throw std::logic_error("scheduled method should take one argument");
        }
    }
};

} // namespace script
} // namespace bipscript

#endif // SCRIPT_H
